/**********************************************************************
 * Copyright (c) 2006 IBM Corporation and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * IBM - Initial API and implementation
 **********************************************************************/
package org.eclipse.cosmos.rm.validation.internal.artifacts;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.eclipse.cosmos.rm.validation.internal.databuilders.DataBuilderRegistry;
import org.eclipse.cosmos.rm.validation.internal.databuilders.IDataBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.TypeInheritanceDataBuilderImpl;

/**
 * A document node is used to represent an SML-IF document 
 * 
 * @author Ali Mehregani
 */
public class DocumentNode
{
	/**
	 * The aliases of this document
	 */
	private List aliases;
	
	/**
	 * The edges of this document node are indexed by reference
	 * type.
	 * 
	 * KEY = A {@link String} indicating the namespace + reference type
	 * VALUE = A {@link List} storing all edges of type KEY.
	 */
	private Map edges;
	
	
	public DocumentNode()
	{
		edges = new Hashtable();
	}
	
	/**
	 * Returns the aliases of this document
	 * 
	 * @return The aliases of this document node
	 */
	public String[] getAliases()
	{
		if (aliases == null)
			return null;
		
		return (String[])aliases.toArray(new String[aliases.size()]);
	}

	
	/**
	 * Add an alias for this document node.  An alias is used to reference
	 * a document.
	 * 
	 * @param alias The alias to be added
	 */
	public void addAlias(String alias)
	{
		if (aliases == null)
			aliases = new ArrayList();
		
		aliases.add(alias);
	}
	
	
	/**
	 * Adds an element edge to this document
	 * 
	 * @param edge The element edge
	 */
	public void addEdge(ElementEdge edge)
	{
		if (edge == null || edge.getReferenceType() == null)
			return;
		
		if (edges == null)
			edges = new Hashtable();
		
		String key = edge.getReferenceNameSpace();
		key = key == null ? edge.getReferenceType() : key + edge.getReferenceType();
		List referenceEdges = (List)edges.get(key);
		if (referenceEdges == null)
			referenceEdges = new ArrayList();
		
		referenceEdges.add(edge);
		edges.put(key, referenceEdges);
	}
	
		
	/**
	 * Return the edges for the reference type passed in.  A valid inheritence tree
	 * must be registered with the data builder for the edges derived from referenceType 
	 * to be returned.
	 * 
	 * @param referenceType The namespace + reference type
	 * @return The edges corresponding to the reference type passed in
	 */
	public ElementEdge[] getEdges(String referenceType)
	{
		IDataBuilder dataBuilder = DataBuilderRegistry.instance().getDataStructureBuilder(TypeInheritanceDataBuilderImpl.ID);
		Map inheritanceMap = dataBuilder == null ? null : (Map)dataBuilder.getDataStructure();
		
		/* Create an empty map */
		if (inheritanceMap == null)
			inheritanceMap = new Hashtable();
		
		List referenceEdges = new ArrayList();
		for (Iterator types = edges.keySet().iterator(); types.hasNext();)
		{
			String currentReference = (String) types.next();
			if (TypeInheritanceDataBuilderImpl.isDerivedType(inheritanceMap, referenceType, currentReference))
			{
				referenceEdges.addAll((List)edges.get(currentReference));
			}			
		}
		
		return (ElementEdge[])referenceEdges.toArray(new ElementEdge[referenceEdges.size()]);
	}

}
