/**********************************************************************
 * Copyright (c) 2006 IBM Corporation and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * IBM - Initial API and implementation
 **********************************************************************/
package org.eclipse.cosmos.rm.validation.internal.reference;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.Hashtable;
import java.util.Map;

import org.eclipse.cosmos.rm.validation.internal.common.IValidationConstants;
import org.eclipse.cosmos.rm.validation.internal.common.SMLCommonUtil;
import org.eclipse.cosmos.rm.validation.internal.common.SMLMessages;
import org.eclipse.cosmos.rm.validation.internal.databuilders.IdentityDataBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.RootElementAliasLookupBuilder;


/**
 * Represents an URI as specified by the grammer: 
 * {@link http://www.ietf.org/rfc/rfc3986.txt}
 * 
 * @author Ali Mehregani
 */
public class URIReference implements IReferenceExpression
{	
	/**
	 * The base URI
	 */
	private URI base;
	
	/**
	 * The reference URI.
	 */
	private URI reference;
	
	/**
	 * The document reference
	 */
	private String documentReference;

	/**
	 * Indicates whether this reference has already been
	 * transformed.
	 */
	private boolean transformed;
	
	
	/**
	 * Contains the aliases of the documents
	 * 
	 * KEY = alias names
	 * VALUE = root element name
	 */
	private Map aliases;
	
	
	/**
	 * Constructor.  The reference must be a valid XPointer expression.
	 * 
	 * @param reference The string representing the XPointer expression 
	 * @throws URISyntaxException If the URI has an incorrect syntax
	 */
	public URIReference(String reference) throws URISyntaxException
	{
		if (reference == null)
			throw new URISyntaxException(null, SMLMessages.errorReferenceNullURI);
		
		/* Trim the white spaces and take out line separators */
		reference = trim(reference);
		
		this.reference = new URI(reference);
	}


	private String trim(String str)
	{
		str = str.trim();
		str = str.replaceAll("\\n", "");
		str = str.replaceAll("\\r", "");
		return str;
	}
	
	
	/**
	 * The document reference is basically the <scheme>+<authority>+<path> 
	 * of the URI.
	 * 
	 * @return the documentAlias
	 * @throws URISyntaxException 
	 */
	public String getDocumentReference() throws URISyntaxException
	{		
		if (documentReference == null)
		{
			String query = reference.getQuery();
			String referenceStr = reference.toString();
			if (query != null && query.length() > 0)
			{				
				referenceStr = referenceStr.substring(referenceStr.indexOf(query));				
			}
			documentReference = referenceStr;
			
			if (aliases == null)
			{
				aliases = (Map)SMLCommonUtil.retrieveDataStructure(RootElementAliasLookupBuilder.ID);
				aliases = aliases == null ? new Hashtable() : aliases;
			}
			
			/* Transform the URI if the aliases cannot be found */
			if (aliases.get(documentReference) == null && !isTransformed())
			{
				documentReference = null;
				transform();
				return getDocumentReference();
			}
		}
		
		return documentReference;
	}


	/**
	 * @see org.eclipse.cosmos.rm.validation.internal.reference.IReferenceExpression#isTransformed()
	 */
	public boolean isTransformed()
	{
		return transformed;
	}


	/**
	 * The transformation will be based on [RFC 3986]: {@link http://www.ietf.org/rfc/rfc3986.txt}.
	 * See section 5 for more details.
	 * 
	 * @throws URISyntaxException If the base URI is invalid 
	 * @see org.eclipse.cosmos.rm.validation.internal.reference.IReferenceExpression#transform()
	 */
	public void transform() throws URISyntaxException
	{		
		transformed = true;
		if (base == null)
		{
			String baseURI = IdentityDataBuilder.retrieveBaseURI();
			if (baseURI == null)
				throw new URISyntaxException(baseURI, SMLMessages.errorBaseNullURI);
			base = new URI(baseURI);
		}
		
		String scheme, authority, path, query, fragment;
		
		/* Determine the base URI */
		if (defined(reference.getScheme())) 
		{
			scheme = reference.getScheme();
			authority = reference.getAuthority();
			path = reference.getPath();
			query = reference.getQuery();			
		}
		else
		{
			if (defined(reference.getAuthority()))
			{
				authority = reference.getAuthority();
				path = reference.getPath();
				query = reference.getQuery();
			}
			else
			{
				if (!defined(reference.getPath()))
				{
					path = base.getPath();
					query = defined(reference.getQuery()) ? reference.getQuery() : base.getQuery();					
				}
				else
				{
					path = reference.getPath().startsWith(IValidationConstants.FORWARD_SLASH) ? 
							removeDotSegments(reference.getPath()) : removeDotSegments(mergePath(base, reference));
					query = reference.getQuery();
				}
				authority = base.getAuthority();				
			}
			scheme = base.getScheme();
		}
		fragment = reference.getFragment();
		
		try
		{
			reference = new URI(scheme, authority, path, query, fragment);
		} 
		catch (URISyntaxException e)
		{
			/* This should never happend */
		}
	}


	/**
	 * See page 32 of {@link http://www.ietf.org/rfc/rfc3986.txt}
	 * 
	 * @param base The base URI
	 * @param relative The relative uri
	 * @return Merged path of the base and the relative URI
	 */
	private String mergePath(URI base, URI relative)
	{
		if (defined(base.getAuthority()) && !defined(base.getPath()))
		{
			return IValidationConstants.FORWARD_SLASH + relative.getPath();
		}
		else
		{
			String basePath = base.getPath();
			int rightMostSlash = basePath.lastIndexOf(IValidationConstants.FORWARD_SLASH);
			basePath = rightMostSlash > 0 ? basePath.substring(0, rightMostSlash + 1) : IValidationConstants.EMPTY_STRING;
			return basePath + relative.getPath();
		}		
	}


	/**
	 * See page 33 of {@link http://www.ietf.org/rfc/rfc3986.txt} 
	 * 
	 * @param input The input path
	 * @return removes the dot segments based on the RFC above
	 */
	private String removeDotSegments(String input)
	{
		String output = "";
		String matchedPrefix;
		
		while (input.length() > 0)
		{
			/* If the input buffer begins with a prefix of "../" or "./",
	           then remove that prefix from the input buffer; otherwise*/
			if ((matchedPrefix = findMatch(input, new String[]{"../", "./"}, new String[0])) != null) 
			{
				int length = matchedPrefix.length();
				input = input.length() > length ? input.substring(length) : IValidationConstants.EMPTY_STRING;
			}
			
			/* if the input buffer begins with a prefix of "/./" or "/.",
	           where "." is a complete path segment, then replace that
	           prefix with "/" in the input buffer; otherwise, */
			else if ((matchedPrefix = findMatch(input, new String[]{"/./"}, new String[]{"/."})) != null) 	
			{
				int length = matchedPrefix.length();
				input = input.length() > length ? input.substring(length) : IValidationConstants.EMPTY_STRING;
				input = IValidationConstants.FORWARD_SLASH + input;
			}

			
			/* if the input buffer begins with a prefix of "/../" or "/..",
			   where ".." is a complete path segment, then replace that
           	   prefix with "/" in the input buffer and remove the last
           	   segment and its preceding "/" (if any) from the output
           	   buffer; otherwise, */
			else if ((matchedPrefix = findMatch(input, new String[]{"/../"}, new String[]{"/.."})) != null) 	
			{
				int length = matchedPrefix.length();
				input = input.length() > length ? input.substring(length) : IValidationConstants.EMPTY_STRING;
				input = IValidationConstants.FORWARD_SLASH + input;
				
				int inx = output.lastIndexOf(IValidationConstants.FORWARD_SLASH);
				output = inx > 0 ? output.substring(0, inx) : IValidationConstants.EMPTY_STRING;  				
			}
			
			/* if the input buffer consists only of "." or "..", then remove
           	   that from the input buffer; otherwise, */
			else if ((matchedPrefix = findMatch(input, new String[0], new String[]{".", ".."})) != null)
			{
				input = IValidationConstants.EMPTY_STRING;				
			}
			
			/* move the first path segment in the input buffer to the end of
           	   the output buffer, including the initial "/" character (if
           	   any) and any subsequent characters up to, but not including,
           	   the next "/" character or the end of the input buffer. */
			else
			{
				int inx = input.indexOf(IValidationConstants.FORWARD_SLASH);
				
				if (inx == 0)
				{
					output += IValidationConstants.FORWARD_SLASH;
					input = input.substring(1);
					inx = input.indexOf(IValidationConstants.FORWARD_SLASH);
				}
					
				if (inx > 0)
				{
					output += input.substring(0, inx);
					input = input.substring(inx);
				}
				else
				{
					output += input;
					input = IValidationConstants.EMPTY_STRING;
				}				 
			}
			
		}
		
		return output;
	}


	private String findMatch(String input, String[] prefix, String[] equalityCheck)
	{
		for (int i = 0; i < prefix.length; i++)
		{
			if (input.startsWith(prefix[i]))
				return prefix[i];
		}
		
		for (int i = 0; i < equalityCheck.length; i++)
		{
			if (input.equals(equalityCheck[i]))
				return equalityCheck[i];
		}
		
		return null;
	}


	private boolean defined(String field)
	{
		return field != null && field.length() > 0;
	}
}
